import re
import typing

from starlette.requests import HTTPConnection

from starlette_session.middleware.authorizationbackends import (
    AuthorizationBackendInterface,
)


class AuthorizationHeaderAuthorizationBackend(AuthorizationBackendInterface):
    def __init__(self, type: str = "Bearer"):
        self.pattern = re.compile(f"^{type.lower()} (.*)$", re.I)

    def get_token(self, connection: HTTPConnection) -> typing.Union[str, None]:
        header = "Authorization"
        if header not in connection.headers:
            return None
        value = connection.headers[header]
        match = self.pattern.match(value)
        return str(match[1]) if match is not None else None



import typing

import jwt
from jwt.exceptions import InvalidTokenError
from starlette.datastructures import Secret

from starlette_session.middleware.codecbackends import (
    CodecBackendInterface,
    DecodeError,
)


class JwtCodecBackend(CodecBackendInterface):
    """
    A codec backend for encoding and decoding JWTs.

    Methods
    -------
    decode(value: str)
        Decodes the provided JWT.
    encode(value: dict[str, Any])
        Encodes the provided claims as a JWT.
    """

    def __init__(
        self,
        key: typing.Union[str, Secret],
        algorithm: str,
    ):
        """
        Constructs an instance of a JwtCodecBackend.

        Parameters
        ----------
        key: Union[str, Secret]
            The key used to sign and verify the integrety of the JWT.
        algorithm: str
            The algorithm used to encode and decode the JWT.
        """
        self._key = key
        self._algorithm = algorithm

    def decode(self, value: str) -> typing.Any:
        """
        Verifies the validity, integrity and decodes the provided JWT.

        raises DecodeError if the provided JWT is not valid or has been tampered with.

        Parameters
        ----------
        value: str
            The JWT that is to be decoded

        Returns
        -------
        dict[str, Any]
        """
        try:
            return jwt.decode(value, key=self._key, algorithms=[self._algorithm])
        except InvalidTokenError as ex:
            raise DecodeError("Failed to decode value") from ex

    def encode(self, data: dict[str, typing.Any]) -> str:
        """
        Encodes the provided claims as a JWT.

        Parameters
        ----------
        data: dict[str, Any]
            The claims that are to be encoded as a JWT.

        Returns
        -------
        str
        """
        return str(jwt.encode(data, key=self._key, algorithm=self._algorithm))



import typing

from starlette.requests import HTTPConnection

from starlette_session.middleware.authorizationbackends import (
    AuthorizationBackendInterface,
)


class CookieAuthorizationBackend(AuthorizationBackendInterface):
    def __init__(self, session_cookie: str = "session"):
        self.session_cookie = session_cookie

    def get_token(self, connection: HTTPConnection) -> typing.Union[str, None]:
        if self.session_cookie not in connection.cookies:
            return None
        return str(connection.cookies[self.session_cookie])



import typing

from starlette.datastructures import MutableHeaders
from starlette.requests import HTTPConnection
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from starlette_session.middleware.authorizationbackends import (
    AuthorizationBackendInterface,
)
from starlette_session.middleware.codecbackends import (
    CodecBackendInterface,
    DecodeError,
)
from starlette_session.middleware.storagebackends import StorageBackendInterface


class SessionMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        codec_backend: CodecBackendInterface,
        storage_backend: StorageBackendInterface,
        authorization_backend: AuthorizationBackendInterface,
    ) -> None:
        self.app = app
        self.codec_backend = codec_backend
        self.storage_backend = storage_backend
        self.authorization_backend = authorization_backend

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):  # pragma: no cover
            await self.app(scope, receive, send)
            return

        connection = HTTPConnection(scope)
        initial_session_was_empty = True

        data = self.authorization_backend.get_token(connection)
        if data is not None:
            try:
                scope["session"] = self.codec_backend.decode(data)
                initial_session_was_empty = False
            except DecodeError:
                scope["session"] = {}
        else:
            scope["session"] = {}

        async def send_wrapper(message: Message) -> None:
            if message["type"] == "http.response.start":
                if scope["session"]:
                    # We have session data to persist.
                    data = self.codec_backend.encode(scope["session"])
                    self.storage_backend.persist(message, data)
                elif not initial_session_was_empty:
                    # The session has been cleared.
                    self.storage_backend.clear(message)
            await send(message)

        await self.app(scope, receive, send_wrapper)



import typing

from starlette.datastructures import MutableHeaders
from starlette.types import Message

from starlette_session.middleware.storagebackends import StorageBackendInterface


class CookieStorageBackend(StorageBackendInterface):
    def __init__(
        self,
        max_age: typing.Optional[int] = 14 * 24 * 60 * 60,
        session_cookie: str = "session",
        path: str = "/",
        same_site: typing.Literal["lax", "strict", "none"] = "lax",
        https_only: bool = False,
    ):
        self.max_age = max_age
        self.session_cookie = session_cookie
        self.path = path
        self.security_flags = "httponly; samesite=" + same_site
        if https_only:  # Secure flag can be used with HTTPS only
            self.security_flags += "; secure"

    def persist(self, message: Message, data: str) -> None:
        headers = MutableHeaders(scope=message)
        header_value = (
            "{session_cookie}={data}; path={path}; {max_age}{security_flags}".format(
                session_cookie=self.session_cookie,
                data=data,
                path=self.path,
                max_age=f"Max-Age={self.max_age}; " if self.max_age else "",
                security_flags=self.security_flags,
            )
        )
        headers.append("Set-Cookie", header_value)

    def clear(self, message: Message) -> None:
        headers = MutableHeaders(scope=message)
        header_value = (
            "{session_cookie}={data}; path={path}; {expires}{security_flags}".format(
                session_cookie=self.session_cookie,
                data="null",
                path=self.path,
                expires="expires=Thu, 01 Jan 1970 00:00:00 GMT; ",
                security_flags=self.security_flags,
            )
        )
        headers.append("Set-Cookie", header_value)



import json
import typing
from base64 import b64decode, b64encode

import itsdangerous
from itsdangerous.exc import BadSignature
from starlette.datastructures import Secret

from starlette_session.middleware.codecbackends import (
    CodecBackendInterface,
    DecodeError,
)


class SignerCodecBackend(CodecBackendInterface):
    """
    A codec backend for encoding and decoding signed tokens.

    Methods
    -------
    decode(value: str)
        Decodes the provided token.
    encode(value: dict[str, Any])
        Encodes the provided dict as a signed token.
    """

    def __init__(
        self,
        key: typing.Union[str, Secret],
        max_age: typing.Optional[int] = 14 * 24 * 60 * 60,
    ):
        """
        Constructs an instance of a SignerCodecBackend.

        Parameters
        ----------
        key: Union[str, Secret]
            The key used to sign and verify the integrety of the token.
        max_age: Optional[int]
            The maximum allowable age of the token. Defaults to 1,209,600 seconds (14 days).
        """
        self.signer = itsdangerous.TimestampSigner(str(key))
        self.max_age = max_age

    def decode(self, value: str) -> typing.Any:
        """
        Verifies the validity and decodes the provided token.

        raises DecodeError if the provided token is not valid.

        Parameters
        ----------
        value: str
            The token that is to be decoded

        Returns
        -------
        dict[str, Any]
        """
        try:
            data = self.signer.unsign(value, max_age=self.max_age)
            return json.loads(b64decode(data))
        except BadSignature as ex:
            raise DecodeError("Failed to decode value") from ex

    def encode(self, data: dict[str, typing.Any]) -> str:
        """
        Encodes the provided data as a token.

        Parameters
        ----------
        data: dict[str, Any]
            The data that is to be encoded as a token.

        Returns
        -------
        str
        """
        base64_encoded_data = b64encode(json.dumps(data).encode("utf-8"))
        return str(self.signer.sign(base64_encoded_data).decode("utf-8"))



import abc
import typing

from starlette.requests import HTTPConnection


class AuthorizationBackendInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, __subclass: type) -> bool:
        return hasattr(__subclass, "get_token") and callable(__subclass.get_token)

    @abc.abstractmethod
    def get_token(self, connection: HTTPConnection) -> typing.Union[str, None]:
        raise NotImplementedError



import abc

from starlette.types import Message


class StorageBackendInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, __subclass: type) -> bool:
        return (
            hasattr(__subclass, "persist")
            and callable(__subclass.persist)
            and hasattr(__subclass, "clear")
            and callable(__subclass.clear)
            or NotImplemented
        )

    @abc.abstractmethod
    def persist(self, message: Message, data: str) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def clear(self, message: Message) -> None:
        raise NotImplementedError



import abc
import typing


class CodecBackendInterface(metaclass=abc.ABCMeta):
    """
    An interface that can be used to implement a codec backend.

    Methods
    -------
    decode(value: str)
        Decodes the provided string.
    encode(value: dict[str, Any])
        Encodes the provided dictionary.
    """

    @classmethod
    def __subclasshook__(cls, __subclass: type) -> bool:
        return (
            hasattr(__subclass, "decode")
            and callable(__subclass.decode)
            and hasattr(__subclass, "encode")
            and callable(__subclass.encode)
            or NotImplemented
        )

    @abc.abstractmethod
    def decode(self, value: str) -> typing.Any:
        """
        Raises NotImplementedError

        Parameters
        ----------
        value: str
            The string to be decoded.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def encode(self, value: dict[str, typing.Any]) -> str:
        """
        Raises NotImplementedError

        Parameters
        ----------
        value: dict[str, Any]
            The value to be encoded.
        """
        raise NotImplementedError


class DecodeError(Exception):
    """
    An error to raise if decoding a token can not or should not happen.
    """

    pass
